// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Per draft-ietf-ipngwg-bsd-api-new-01.txt.
//
// This file contains IPv6 specific information for use
// by Winsock2 compatible applications.
//
// It also declares functionality exported from wship6.lib,
// an application helper library for IPv6.
//

#ifndef WS2IP6_INCLUDED
#define WS2IP6_INCLUDED 1

#include <ip6exp.h>
#include <devioctl.h>
#include <ntddip6.h>

//
// Define INLINE if it isn't already defined.
//
#ifndef INLINE
#ifdef _MSC_VER
#define INLINE __inline
#else
#define INLINE extern inline /* GNU style */
#endif
#endif

typedef struct in6_addr IN6_ADDR;
typedef struct in6_addr *PIN6_ADDR;
typedef struct in6_addr FAR *LPIN6_ADDR;


struct sockaddr_in6 {
    short sin6_family;
    u_short sin6_port;
    u_long sin6_flowinfo;
    struct in6_addr sin6_addr;
};

typedef struct sockaddr_in6 SOCKADDR_IN6;
typedef struct sockaddr_in6 *PSOCKADDR_IN6;
typedef struct sockaddr_in6 FAR *LPSOCKADDR_IN6;


INLINE int IN6_IS_ADDR_UNSPECIFIED(const struct in6_addr *a)
{
    return ((a->s6_dwords[0] == 0) &&
            (a->s6_dwords[1] == 0) &&
            (a->s6_dwords[2] == 0) &&
            (a->s6_dwords[3] == 0));
}
 
INLINE int IN6_IS_ADDR_LOOPBACK(const struct in6_addr *a)
{
    return ((a->s6_dwords[0] == 0) &&
            (a->s6_dwords[1] == 0) &&
            (a->s6_dwords[2] == 0) &&
            (a->s6_dwords[3] == htonl(1)));
}
 
INLINE int IN6_IS_ADDR_LINKLOCAL(const struct in6_addr *a)
{
    return ((a->s6_bytes[0] == 0xfe) &&
            ((a->s6_bytes[1] & 0xc0) == 0x80));
}

INLINE int IN6_ADDR_EQUAL(const struct in6_addr *a, const struct in6_addr *b)
{
    return (memcmp(a, b, sizeof(struct in6_addr)) == 0);
}


#define IPPROTO_IPV6            41

//
// Socket options at the IPPROTO_IPV6 level.
//
#define	IPV6_MULTICAST_IF       9  // set/get IP multicast i/f
#define	IPV6_MULTICAST_HOPS     10 // set/get IP multicast ttl
#define	IPV6_MULTICAST_LOOP     11 // set/get IP multicast loopback
#define	IPV6_ADD_MEMBERSHIP     12 // add an IP group membership
#define	IPV6_DROP_MEMBERSHIP    13 // drop an IP group membership

//
// Definitions for exports from wship6.lib.
//

extern const char * WSAAPI
GetErrorString(IP_STATUS Error);

extern char * WSAAPI
inet6_ntoa(const struct in6_addr *Address);

extern int WSAAPI
inet6_addr(const char *s, struct in6_addr *Address);

extern struct hostent * WSAAPI
getnodebyname(const char *name, int af, int flags);

// Flag values are not implemented yet.
#define AI_DEFAULT      0

extern void WSAAPI
freehostent(struct hostent *ptr);

extern const struct in6_addr in6addr_any;
extern const struct in6_addr in6addr_loopback;

#endif // WS2IP6_INCLUDED
