// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Performs socket calls to add & drop multicast groups.
//

#include <winsock2.h>
#include <ws2ip6.h>
#include <stdio.h>
#include <stdlib.h>

int GetAddress(char *astr, struct in6_addr *address);
int one = 1;

void SendBSD(struct in6_addr Group, char *port, char *IF);
void ReceiveBSD(struct in6_addr Group, char *port, char *IF);
void SendWSA(struct in6_addr Group, char *port, char *IF);
void ReceiveWSA(struct in6_addr Group, char *port, char *IF);



#define BUFSIZE 1600

const struct in6_addr in6addr_any = { 0 };

int Loopback = FALSE;

void
usage(void)
{
    printf("usage: mldtest s|r group-addr port interface# [loop [w]]\n");
    printf("usage: s = act as sender, r = act as receiver\n");
    printf("usage: w = use winsock2 style functions instead of bsd functions\n");
    exit(1);
}


int __cdecl main(int argc, char **argv)
{
    int error;
    int i;
    WSADATA WsaData;
    char buffer[BUFSIZE];
    int parm;

    SOCKET bsd, bsdsend;
	struct in6_addr GroupAddr;
	struct ipv6_mreq Query;
    struct sockaddr_in6 SA;
    DWORD  bytesout;

    int WSAProtoIds[] = { IPPROTO_UDP, 0};
    LPWSAPROTOCOL_INFO ProtoInfo;
    DWORD Bufsize = BUFSIZE;
    DWORD WSAFlags;
    SOCKET wsa;
    WSABUF CalleeData;

    if (argc < 5) {
        usage();
    }


    // initialize winsock
    error = WSAStartup( MAKEWORD(2, 0), &WsaData );
    if ( error ) {
        printf("ttcp6: WSAStartup failed %ld:", WSAGetLastError());
    }

    // convert the comand-line address to the internal format
	if (GetAddress(argv[2], &GroupAddr) == FALSE) {
		printf("bad address supplied!\n");
		exit(1);
	}

    // loopback?
    if (argc >5 && argv[5][0] == 'y')
        Loopback = TRUE;


    // which socket style to use?
    if (argc >6 && argv[6][0] == 'w')
        goto Winsock;

    // BSD-style
    if (argv[1][0] == 's')
        SendBSD(GroupAddr, argv[3], argv[4]);
    else
        ReceiveBSD(GroupAddr, argv[3], argv[4]);

    WSACleanup();
    return 0;


Winsock:

    if (argv[1][0] == 's')
        SendWSA(GroupAddr, argv[3], argv[4]);
    else
        ReceiveWSA(GroupAddr, argv[3], argv[4]);

    WSACleanup();
    return 0;

}



int
GetAddress(char *astr, struct in6_addr *address)
{
    struct sockaddr_in6 sin;
    int addrlen = sizeof sin;

    sin.sin6_family = AF_INET; // shouldn't be required but is

    if ((WSAStringToAddress(astr, AF_INET6, NULL,
                           (struct sockaddr *)&sin, &addrlen)
                    == SOCKET_ERROR) ||
        (sin.sin6_port != 0))
        return FALSE;

    // The user gave us a numeric IP address.

    memcpy(address, &sin.sin6_addr, sizeof *address);
    return TRUE;
}



// send data using a bsd-style socket.
void SendBSD(struct in6_addr Group, char *port, char *IF)
{
    int error, parm, i;
    SOCKET bsdsend;
    struct sockaddr_in6 SA;
    char sendbuf[10];

    // Send some data to the group
    bsdsend = socket(AF_INET6, SOCK_DGRAM, 0);
    if (bsdsend == INVALID_SOCKET) {
        printf("socket() failed: %ld\n", GetLastError( ) );
        exit(1);    
    }

    if (setsockopt(bsdsend, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one))) {
        printf("setsockopt() failed!\n");
        exit(1);
    }

    // bind the socket to a local address & port
    SA.sin6_addr = in6addr_any;
    SA.sin6_port = 0;
    SA.sin6_flowinfo = 0;
    SA.sin6_family = AF_INET6;
    if (bind(bsdsend, (struct sockaddr *)&SA, sizeof(SA))) {
 	    printf("bind failed: %d\n",WSAGetLastError());
    	exit(1);
    }


    SA.sin6_addr = Group;
    SA.sin6_port = htons((short)atoi(port));
    error = connect(bsdsend, (struct sockaddr *)&SA, sizeof(SA));
    if (error) {
         printf("connect() failed: %d\n",WSAGetLastError());
       exit(1);
    }

    
    // set the default outbound interface.  For simplicity we use the
    // same if as we receive on, but that need not be the case.
    parm = atoi(IF);
    error = setsockopt(bsdsend, IPPROTO_IPV6, IPV6_MULTICAST_IF, (char *)&parm, sizeof(int));
    if (error) {
        printf("setsockopt() failed: %d\n",WSAGetLastError());
        exit(1);
    }

    // Set the hop limit for multicast packets we send.
    parm = 3;
    error = setsockopt(bsdsend, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, (char *)&parm, sizeof(int));
    if (error) {
        printf("setsockopt() failed: %d\n",WSAGetLastError());
        exit(1);
    }

    // Set loopback.
    if (Loopback)
        parm = 1;
    else
        parm = 0;
    error = setsockopt(bsdsend, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, (char *)&parm, sizeof(int));
    if (error) {
        printf("set loopback option failed: %d\n",WSAGetLastError());
    }


    for (i=0; i<1000; i++) {
        sprintf(sendbuf,"%4d-bsd",i);
        send(bsdsend, (const char FAR *)sendbuf, 10, 0);
        Sleep(250); 
    }

    closesocket(bsdsend);

}


void ReceiveBSD(struct in6_addr Group, char *port, char *IF)
{

    int error, i, bytes;
    char buffer[BUFSIZE];
    SOCKET bsd;
	struct ipv6_mreq Query;
    struct sockaddr_in6 SA, RSA;
    int FromLen;


        // open a socket to use
    bsd = socket(AF_INET6, SOCK_DGRAM, 0);
    if (bsd == INVALID_SOCKET) {
        printf("socket() failed: %ld\n", GetLastError( ) );
        exit(1);    
    }

    if (setsockopt(bsd, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one))) {
        printf("setsockopt() failed!\n");
        exit(1);
    }

    // bind the socket to the multicast address & port
    SA.sin6_port = htons((short)atoi(port));
    SA.sin6_addr = Group;
    SA.sin6_family = AF_INET6;
    if (bind(bsd, (struct sockaddr *)&SA, sizeof(SA))) {
        // bind to group address failed, try generic address.
        SA.sin6_addr = in6addr_any;
        if (bind(bsd, (struct sockaddr *)&SA, sizeof(SA))) {
 		    printf("bind failed: %d\n",WSAGetLastError());
		    exit(1);
        }
    }

    // notify IP of our interest in this group
	Query.ipv6mr_multiaddr = Group;
	Query.ipv6mr_interface = atoi(IF);
    error = setsockopt(bsd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, (char *)&Query, sizeof Query);
    if (error) {
        printf("setsockopt() failed: %d\n",WSAGetLastError());
        exit(1);
    }

    // read some data from the socket
    for (i=0; i<1000; i++) {
        FromLen = sizeof(RSA);
        bytes = recvfrom(bsd, buffer, BUFSIZE, 0, (struct sockaddr *)&RSA, &FromLen);
        if (bytes != SOCKET_ERROR) {
            buffer[bytes] = '\0';
        printf("%d, %s\n",i,buffer);
        } else {
           error = WSAGetLastError();
           printf("receiveBSD: recv error %d\n",error);
        }
    }
    
    // OK we're done
    closesocket(bsd);
    return;
}

void SendWSA(struct in6_addr Group, char *port, char *IF)
{
    int error, parm, i;
    char buffer[BUFSIZE];
    struct sockaddr_in6 SA;
    DWORD  bytesout;
    int WSAProtoIds[] = { IPPROTO_UDP, 0};
    LPWSAPROTOCOL_INFO ProtoInfo;
    DWORD Bufsize = BUFSIZE;
    DWORD WSAFlags;
    SOCKET wsa, wsasend;
    WSABUF CalleeData;
    char sendbuf[10];

    // Get the protocol info structure for UDP service
    error = WSAEnumProtocols(WSAProtoIds, (LPWSAPROTOCOL_INFO) buffer, &Bufsize);
    if (error == SOCKET_ERROR) {
        printf("WSAEnumProtocols() failed: %d\n",WSAGetLastError());
        exit(1);
    }
    ProtoInfo = (LPWSAPROTOCOL_INFO) buffer;

    // open a WSA socket
    WSAFlags = WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF;
    wsa = WSASocket(AF_INET6, SOCK_DGRAM, 0, ProtoInfo, 0, WSAFlags);
    if (wsa == INVALID_SOCKET) {
        printf("WSASocket() failed: %d\n",WSAGetLastError());
        exit(1);
    }

    if (setsockopt(wsa, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one))) {
        printf("setsockopt() failed!\n");
        exit(1);
    }



    // Join the multicast group
    SA.sin6_addr = Group;
    WSAHtons(wsa, (short)atoi(port), &SA.sin6_port);
    SA.sin6_flowinfo = 0;
    SA.sin6_family = AF_INET6;
    WSAFlags = JL_SENDER_ONLY;
    CalleeData.len = 0;
    CalleeData.buf = NULL;
    wsasend = WSAJoinLeaf(wsa, (struct sockaddr FAR *)&SA, sizeof(SA), NULL, &CalleeData, NULL, NULL, WSAFlags);
    if (wsasend == INVALID_SOCKET) {
        printf("WSAJoinLeaf() failed: %d\n",WSAGetLastError());
        exit(1);
    }
   
    // connect to group address for sending.
    error = WSAConnect ( wsasend, (struct sockaddr FAR *)&SA, sizeof(SA), NULL, &CalleeData, NULL, NULL);
    if (error == SOCKET_ERROR) {
        printf("WSAConnect() failed: %d\n",WSAGetLastError());
        exit(1);
    }
 

    // Set the outbound multicast default ttl
    parm = 3;
    error = WSAIoctl( wsasend, SIO_MULTICAST_SCOPE, (LPVOID)&parm, sizeof(parm), (LPVOID) buffer,
        BUFSIZE , (LPDWORD) &bytesout, NULL, NULL);
     if (error == SOCKET_ERROR) {
        printf("WSAIoctl() failed: %d\n",WSAGetLastError());
        exit(1);
    }
 

    // send some data
    for (i=0; i<1000 ; i++) {
        sprintf(sendbuf,"%4d wsa",i);
        send(wsasend, (const char FAR *)sendbuf, 10, 0);
        Sleep(250); 
    }

    
    // OK we're done
    closesocket(wsa);
    closesocket(wsasend);
    return;

}



void ReceiveWSA(struct in6_addr Group, char *port, char *IF)
{
    int error, i, status;
    char buffer[BUFSIZE];
    int parm;
	struct in6_addr GroupAddr;
    struct sockaddr_in6 SA, RSA;
    int FromLen;
    DWORD  bytesout;

    int WSAProtoIds[] = { IPPROTO_UDP, 0};
    LPWSAPROTOCOL_INFO ProtoInfo;
    DWORD Bufsize = BUFSIZE;
    DWORD WSAFlags;
    SOCKET wsa, wsatrash;
    WSABUF CalleeData;


    // Get the protocol info structure for UDP service
    error = WSAEnumProtocols(WSAProtoIds, (LPWSAPROTOCOL_INFO) buffer, &Bufsize);
    if (error == SOCKET_ERROR) {
        printf("WSAEnumProtocols() failed: %d\n",WSAGetLastError());
        exit(1);
    }
    ProtoInfo = (LPWSAPROTOCOL_INFO) buffer;

    // open a WSA socket
    WSAFlags = WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF;
    wsa = WSASocket(AF_INET6, SOCK_DGRAM, 0, ProtoInfo, 0, WSAFlags);
    if (wsa == INVALID_SOCKET) {
        printf("WSASocket() failed: %d\n",WSAGetLastError());
        exit(1);
    }

    if (setsockopt(wsa, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one))) {
        printf("setsockopt() failed!\n");
        exit(1);
    }


    // bind the socket to the multicast address & port
    SA.sin6_addr = Group;
    WSAHtons(wsa, (short)atoi(port), &SA.sin6_port);
    SA.sin6_flowinfo = 0;
    SA.sin6_family = AF_INET6;
    if (bind(wsa, (struct sockaddr *)&SA, sizeof(SA))) {
        // bind to group address failed, try generic address.
        SA.sin6_addr = in6addr_any;
        if (bind(wsa, (struct sockaddr *)&SA, sizeof(SA))) {
 		    printf("bind failed: %d\n",WSAGetLastError());
		    exit(1);
        }
    }

    
    // Join the multicast group
    SA.sin6_addr = Group;
    WSAFlags = JL_RECEIVER_ONLY;
    CalleeData.len = 0;
    CalleeData.buf = NULL;
    wsatrash = WSAJoinLeaf(wsa, (struct sockaddr FAR *)&SA, sizeof(SA), NULL, &CalleeData, NULL, NULL, WSAFlags);
    if (wsatrash == INVALID_SOCKET) {
        printf("WSAJoinLeaf() failed: %d\n",WSAGetLastError());
        exit(1);
    }
 

    // read some data from the socket
    for (i=0; i<1000; i++) {
        FromLen = sizeof(RSA);
        CalleeData.len = BUFSIZE;
        CalleeData.buf = buffer;
        status = WSARecvFrom (wsa, &CalleeData, 1, &Bufsize, &WSAFlags, (struct sockaddr *)&RSA, &FromLen, NULL, NULL);
        if (status != SOCKET_ERROR) {
            buffer[Bufsize] = '\0';
        printf("%d, %s\n",i,buffer);
        } else {
           error = WSAGetLastError();
           printf("receiveBSD: recv error %d\n",error);
        }
    }
    
    // OK we're done
    closesocket(wsa);
    closesocket(wsatrash);
    return;
}

