// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1985-1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Send raw IPv6 packet.
//

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <winsock2.h>
#include <ws2ip6.h>

#define IP_VERSION 0x00000060

//
// Network byte-order is big-endian.
// NT runs in little-endian mode on all supported architectures.
//
__inline ushort
net_short(ushort x)
{
    return (((x & 0xff) << 8) | ((x & 0xff00) >> 8));
}

void err(char *s);
int parse_addr(char *s, struct sockaddr_in6 *sin6);

void _CRTAPI1
main(int argc, char **argv)
{
    int Error;
    WSADATA WsaData;
    SOCKET Sock;
	char TempBuf[1000];
	IPv6Header *IP;
	int Count;
	uchar HopLimit, Protocol;
    struct sockaddr_in6 Us, Peer;
    short Port;
    char *Host;

	Count = 29;  // Amount of raw data to send after IPv6 header.
	Port = 4000;  // Arbitrary (and meaningless).
	HopLimit = 10;  // Arbitrary.
	Protocol = 59;  // The "No Next Header" type.

    if (argc <= 2) goto usage;
        
    //
    // Initialize winsock.
    //
    Error = WSAStartup(0x0101, &WsaData);
    if (Error == SOCKET_ERROR) {
        printf("rawsend: WSAStartup failed %ld:", WSAGetLastError());
    }

    //
    // Establish our peer's address.
    //
	memset((char *)&Peer, 0, sizeof(Peer));
	Host = argv[1];
	if (!parse_addr(Host, &Peer))
        err("bad hostname");
	Peer.sin6_port = htons(Port);
	
    //
    // Establish our address.
    //
	memset((char *)&Us, 0, sizeof(Us));
	Host = argv[2];
	if (!parse_addr(Host, &Us))
        err("bad hostname");
	Us.sin6_port = 0;
	Us.sin6_family = AF_INET6;

    //
    // Create a raw IPv6 socket.
    //
	Sock = socket(PF_INET6, SOCK_RAW, 0);
	if (Sock == INVALID_SOCKET) {
        printf("socket() failed: %ld\n", GetLastError( ) );
        exit(1);
	}

    //
    // Bind our source address to the socket.
    // For raw send, this has the effect of picking an outgoing interface.
    //
    if (bind(Sock, (PSOCKADDR)&Us, sizeof(Us)) < 0)
        err("bind");

    //
    // Put an IPv6 header at the beginning of our packet buffer.
    //
    IP = (IPv6Header *)((uchar *)TempBuf);
    IP->VersClassFlow = IP_VERSION;
    IP->NextHeader = Protocol;
    IP->HopLimit = HopLimit;
    IP->Source = Us.sin6_addr;
    IP->Dest = Peer.sin6_addr;
    IP->PayloadLength = net_short((ushort)Count);

    //
    // Send the packet.
    //
	if (sendto(Sock, (char *)TempBuf, sizeof(IPv6Header) + Count, 0,
               (PSOCKADDR)&Peer, sizeof(Peer)) < 0) {
        printf("sendto() failed: %ld\n", GetLastError( ) );
        exit(1);
    }

    //
    // Cleanup.
    //
    closesocket(Sock);
    WSACleanup();
    exit(0);

usage:
    fprintf(stderr, "Usage: rawsend destination source\n");
    WSACleanup();
    exit(1);
}


//
// Print an error message.
//
void err(s)
char *s;
{
        fprintf(stderr,"rawsend: ");
        perror(s);
        fprintf(stderr,"errno=%ld\n", WSAGetLastError( ));
        WSACleanup();
        exit(1);
}


//
// Does not currently support DNS names, but should.
// (If WSAStringToAddress fails, then try DNS.)
//
int
parse_addr(char *s, struct sockaddr_in6 *sin6)
{
    struct hostent *h;

    sin6->sin6_family = AF_INET6;

    if (!inet6_addr(s, &sin6->sin6_addr)) {
        // Presumably the user gave us a DNS name.

        h = getnodebyname(s, AF_INET6, AI_DEFAULT);
        if (h == NULL)
            return FALSE;

        memcpy(&sin6->sin6_addr, h->h_addr, sizeof(struct in6_addr));
        freehostent(h);
    }

    return TRUE;
}
