// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) 1998 Microsoft Corporation
//
// This file is part of the Microsoft Research IPv6 Network Protocol Stack.
// You should have received a copy of the Microsoft End-User License Agreement
// for this software along with this release; see the file "license.txt".
// If not, please see http://www.research.microsoft.com/msripv6/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// IPv6 header option types and headers.
//


#ifndef IP6OPTS_INCLUDED
#define IP6OPTS_INCLUDED 1

//
// Generic Extension Header
// 
typedef struct ExtensionHeader {
    uchar NextHeader;
    uchar HeaderExtLength;  // In 8-byte units, not counting first 8.
} ExtensionHeader;

//
// Generic Options Header 
// 
typedef struct OptionHeader {
    uchar Type;
    uchar DataLength;  // In 8-byte units, not counting first 8.
} OptionHeader;

//
// Multi-byte pad option header.
//
typedef struct IPv6PadNOption {
    uchar Type;        // Must be 1 for PadN option.
    uchar DataLength;  // Length of the data that follows.
} IPv6PadNOption;

//
// Format of the router alert within the Hop-by-Hop Option header.
//
typedef struct IPv6RouterAlertOption {
    uchar Type;
    uchar Length;
    ushort Value;
} IPv6RouterAlertOption;

typedef struct Options {
    uint JumboLength;  // Length of packet excluding IPv6 header.
    uint Alert;        // REVIEW: for now, buffer pointer to the Alert option
} Options;

// Option Header Values
#define OPT6_PAD_1 0            // Single byte pad
#define OPT6_PAD_N 1            // Multiple byte pad
#define OPT6_JUMBO_PAYLOAD 194  // Jumbo payload (greater than 65,535 bytes)
#define OPT6_ROUTER_ALERT 199   // BUGBUG: this is bogus, waiting on IANA

// Type of actions to be taken with unrecognized header options
#define OPT6_ACTION(a)  ((a) & 0xc0)    // get action bits
#define OPT6_A_SKIP     0x00            // Skip and continue
#define OPT6_A_DISCARD  0x40            // Discard packet
#define OPT6_A_SEND_ICMP_ALL  0x80      // Send ICMP regardless of source addr
#define OPT6_A_SEND_ICMP_NON_MULTI 0xc0 // Send ICMP if non-multicast src addr

// Various constants
#define EXT_HDR_LEN  sizeof(ExtensionHeader) 
#define OPT_HDR_LEN  sizeof(OptionHeader)    
#define EXT_LEN_UNIT 8          // 8-bytes units used for extension hdr length
#define DEFAULT_EXT_HDR_LEN 8   // (bytes) minimum/unaccouted length of 
                                // Destination and Hop-by-Hop headers

#define ADJUST_PACKET_PARAMS(pkt, cnt) {pkt->ContigSize -= cnt; \
                                        pkt->TotalSize -= cnt; \
                                        pkt->Position += cnt; }



#endif  // IP6OPTS_INCLUDED





